{
  Saving TOleContainer in RTF file.
  Use GetOleContainerRTFString in OnSaveComponentToFile:

  case SaveFormat of
   ...
   rvsfRTF:
      begin
        if SaveMe is TOleContainer then
          OutStr := GetOleContainerRTFString(TOleContainer(SaveMe))
        ...
      end;
  end;

  You can provide an optional image representing this object in the second
  parameter.

  This function is based on work by David Michael,
  http://www.trichview.com/forums/viewtopic.php?t=108

}

unit SaveOleContainerToRTF;

interface

uses Windows, Classes, Graphics, OleCtnrs, ActiveX, ComObj,
  RVClasses, RVFuncs, RVTypes, RVFMisc, SysUtils;

function GetOleContainerRTFString(OleContainer: TOleContainer;
  Image: TMetafile = nil): String;

implementation

uses Controls;

procedure SaveOleObjectAsDocumentToStream(OleContainer:TOleContainer;
  Stream:TStream);
var
  PersistStorage: IPersistStorage;
  TempLockBytes: ILockBytes;
  TempStorage: IStorage;
  DataHandle: HGlobal;
  BufferSize: integer;
  Buffer: Pointer;
begin
  if OleContainer.OleObjectInterface<>nil then begin
    OleContainer.OleObjectInterface.QueryInterface(IPersistStorage,PersistStorage);
    if PersistStorage<>nil then begin
      OleCheck(CreateILockBytesOnHGlobal(0,True,TempLockBytes));
      OleCheck(StgCreateDocfileOnILockBytes(TempLockBytes,STGM_READWRITE
        or STGM_SHARE_EXCLUSIVE or STGM_CREATE,0,TempStorage));
      OleCheck(OleSave(PersistStorage,TempStorage,False));
      PersistStorage.SaveCompleted(nil);
      OleCheck(GetHGlobalFromILockBytes(TempLockBytes,DataHandle));
      BufferSize:=GlobalSize(DataHandle);
      Buffer:=GlobalLock(DataHandle);
      try
        Stream.WriteBuffer(Buffer^,BufferSize);
      finally
        GlobalUnlock(DataHandle);
      end; // try
    end;
  end;
end;
{------------------------------------------------------------------------------}
// This function is unreliable, so using the next one
function GetOleObjectMetaPict(OleContainer:TOleContainer): HGlobal;
var
  DataObject: IDataObject;
  FormatEtc: TFormatEtc;
  Medium: TStgMedium;
begin
   Result := 0;
   if OleContainer.OleObjectInterface<>nil then begin
     // metapict won't draw unless the object's server is running
     if OleContainer.State=osLoaded then
       OleContainer.Run;
     OleContainer.OleObjectInterface.QueryInterface(IDataObject, DataObject);
     if DataObject<>nil then begin
       if OleContainer.Iconic then begin
         FormatEtc.cfFormat:=CF_ENHMETAFILE;
         FormatEtc.ptd:=nil;
         FormatEtc.dwAspect:=DVASPECT_ICON;
         FormatEtc.lIndex:=-1;
         FormatEtc.tymed:=TYMED_ENHMF;
         if Succeeded(DataObject.GetData(FormatEtc,Medium)) then
           Result:=Medium.hEnhMetaFile;
         end
       else begin
         OleContainer.OleObjectInterface.QueryInterface(IDataObject,DataObject);
         if DataObject <> nil then begin
           FormatEtc.cfFormat:=CF_ENHMETAFILE;
           FormatEtc.ptd:=nil;
           FormatEtc.dwAspect:=DVASPECT_CONTENT;
           FormatEtc.lIndex:=-1;
           FormatEtc.tymed:=TYMED_ENHMF;
           FillChar(Medium,SizeOf(Medium),0);
           if Succeeded(DataObject.GetData(FormatEtc, Medium)) then
             Result:=Medium.hEnhMetaFile;
         end;
       end;
     end;
   end;
end;
{------------------------------------------------------------------------------}
function GetOleObjectMetafile(OleContainer:TOleContainer): TMetafile;
var Canvas: TMetafileCanvas;
begin
  OleContainer.Handle;
  Result := TMetafile.Create;
  try
    Result.Width := OleContainer.Width;
    Result.Height := OleContainer.Height;
    Canvas := TMetafileCanvas.Create(Result, 0);
    try
      OleContainer.PaintTo(Canvas,0,0);
    finally
      Canvas.Free;
    end;
  except
    Result.Free;
    Result := nil;
  end;
end;
{------------------------------------------------------------------------------}
function GetOleContainerRTFString(OleContainer: TOleContainer;
  Image: TMetafile = nil): String;

   function GetTwipsPerPixel:double;
      var
         DC: HDC;
      begin
      DC:=CreateCompatibleDC(0);
      try
         if RichViewPixelsPerInch>0 then
            Result:=(72*20)/RichViewPixelsPerInch
         else
            Result:=(72*20)/GetDeviceCaps(DC,LOGPIXELSY);
      finally
         DeleteDC(DC);
      end; // try
      end;

   function GetLEIntHexStr(int:integer):String;
   var ii: integer;
       bb: byte;
   begin
      // create "little endian" hexadecimal format
     Result:='';
     for ii:=1 to 4 do begin
       bb := int and $FF;
       Result:=Result+IntToHex(bb,2);
       int := int shr 8;
     end;
   end;

   function GetASCIIHexStr(const str: String): String;
   var ss: integer;
   begin
      // creates hexadecimal list from string's characters
     Result:='';
     for ss := 1 to Length(str) do
       Result:=Result+IntToHex(Ord(str[ss]),2);
     Result:=Result+'00';
   end;

   function GetStreamHexStr(stream: TStream): String;
   var
     ss, ssStop, ssRest, cc, idx: integer;
     bb: byte;
     hex: String;
   begin
     // broken into 78-char lines, like TRichEdit's RTF embedded object output
     ssStop := (stream.Size-stream.Position) div 39;
     ssRest := (stream.Size-stream.Position) mod 39;
     cc := ssStop*(78+2);
     if ssRest > 0 then
       inc(cc, ssRest*2+2);
     SetLength(Result, cc);
     idx := 1;
     for ss := 1 to ssStop do begin
       for cc:= 0 to 38 do begin
         stream.Read(bb,1);
         hex := IntToHex(bb,2);
         Result[idx] := hex[1];
         inc(idx);
         Result[idx] := hex[2];
         inc(idx);
       end;
       Result[idx] := #13;
       inc(idx);
       Result[idx] := #10;
       inc(idx);
     end;
     if ssRest>0 then begin
       for cc := 1 to ssRest do begin
         stream.Read(bb,1);
         hex := IntToHex(bb,2);
         Result[idx] := hex[1];
         inc(idx);
         Result[idx] := hex[2];
         inc(idx);
       end;
       Result[idx] := #13;
       inc(idx);
       Result[idx] := #10;
       inc(idx);
     end;
     Assert(idx=Length(Result)+1);
   end;

   const
      MetaPictName: String = 'METAFILEPICT';

   var

      className: String;
      sStream: TStringStream;
      mStream: TRVMemoryStream;
      ii: integer;
      metaFile: TMetafile;
begin
  Result := '';
  if OleContainer.OleObjectInterface=nil then
    exit;
  ClassName := OleContainer.OleClassName;
  sStream := TStringStream.Create('');
  try
    // write data
    sStream.WriteString('{\object\objemb');
    sStream.WriteString('{\*\objclass '+ClassName+'}');
    ii:=Round(OleContainer.Width*GetTwipsPerPixel);
    sStream.WriteString('\objw'+IntToStr(ii));
    ii:=Round(OleContainer.Height*GetTwipsPerPixel);
    sStream.WriteString('\objh'+IntToStr(ii));
    sStream.WriteString('{\*\objdata'#13#10);
    sStream.WriteString('01050000'#13#10);
    sStream.WriteString('02000000'#13#10);
    sStream.WriteString(GetLEIntHexStr(Length(className)+1)+#13#10);
    sStream.WriteString(GetASCIIHexStr(className)+#13#10);
    sStream.WriteString('00000000'#13#10);
    sStream.WriteString('00000000'#13#10);
    mStream:=TRVMemoryStream.Create;
    try
      SaveOleObjectAsDocumentToStream(OleContainer, mStream);
      mStream.Seek(0,0);
      sStream.WriteString(GetLEIntHexStr(mStream.Size)+#13#10);
      sStream.WriteString(GetStreamHexStr(mStream));
    finally
      mStream.Free;
    end; // try
    // write metapict
    sStream.WriteString('01050000'#13#10'05000000'#13#10);
    sStream.WriteString(GetLEIntHexStr(Length(MetaPictName)+1)+#13#10);
    sStream.WriteString(GetASCIIHexStr(MetaPictName)+#13#10);
    metaFile := TMetafile.Create;
    try
      if Image<>nil then
        metaFile.Assign(Image)
      else begin
        metaFile.Free;
        metaFile := GetOleObjectMetafile(OleContainer);
      end;
      //metaFile.Handle := GetOleObjectMetaPict(OleContainer);
      ii:=Round(metaFile.Width*GetTwipsPerPixel);
      sStream.WriteString(GetLEIntHexStr(ii)+#13#10);
      ii:=-Round(metaFile.Height*GetTwipsPerPixel);
      sStream.WriteString(GetLEIntHexStr(ii)+#13#10);
      mStream:=TRVMemoryStream.Create;
      try
        metaFile.Enhanced:=false;
        metaFile.SaveToStream(mStream);
        mStream.Seek(22,0); // sizeof(TMetafileHeader)=22
        sStream.WriteString(GetLEIntHexStr((mStream.Size-22)+8)+#13#10); // sizeof(TMetafileHeader)=22
        ii:=Round(metaFile.Width*GetTwipsPerPixel);
        ii:=(ii shl 16) or $0008;
        sStream.WriteString(GetLEIntHexStr(ii)+#13#10);
        ii:=Round(metaFile.Height*GetTwipsPerPixel);
        sStream.WriteString(GetLEIntHexStr(ii)+#13#10);
        sStream.WriteString(GetStreamHexStr(mStream));
        sStream.WriteString('}');
        // write "result" (metapict again)
        sStream.WriteString('{\result');
        sStream.WriteString('{\pict\wmetafile8');
        ii:=Round(metaFile.Width*GetTwipsPerPixel);
        sStream.WriteString('\picw'+inttostr(ii));
        ii:=Round(metaFile.Height*GetTwipsPerPixel);
        sStream.WriteString('\pich'+inttostr(ii));
        ii:=Round(OleContainer.Width*GetTwipsPerPixel);
        sStream.WriteString('\picwgoal'+inttostr(ii));
        ii:=Round(OleContainer.Height*GetTwipsPerPixel);
        sStream.WriteString('\pichgoal'+inttostr(ii)+#13#10);
        mStream.Seek(22,0); // sizeof(TMetafileHeader)=22
        sStream.WriteString(GetStreamHexStr(mStream));
        sStream.WriteString('}}');
      finally
        mStream.Free;
      end; // try
    finally
      metaFile.Free;
    end; // try
    sStream.WriteString('}');
    Result := sStream.DataString;
  finally
    sStream.Free;
  end; // try
end;


end.
